# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0 Sample
# File Name       : soracom_data_upload_thread.py
# Copyright       : 2015-2020 SORACOM, INC.
# ======================================================

from time import sleep
import threading
from multiprocessing import Pool
import importlib

from logging import getLogger

logger = getLogger(__name__)

DATA_UPLOAD = None
DATA_UPLOAD_MODULE = ''

def UploadDataPoolRun(upload_data_pool_module,
                      device_id, timestamp, data_type, images, data, data_id, resources, options):
    global DATA_UPLOAD
    global DATA_UPLOAD_MODULE
    result = None
    logger.debug("DATA_UPLOAD_MODULE is %s.", DATA_UPLOAD_MODULE)
    if upload_data_pool_module is not None and len(upload_data_pool_module) > 0:
        if DATA_UPLOAD_MODULE != upload_data_pool_module:
            logger.info("import data upload module name is %s.", upload_data_pool_module)
            DATA_UPLOAD = importlib.import_module(upload_data_pool_module)
            if DATA_UPLOAD is not None:
                DATA_UPLOAD_MODULE = upload_data_pool_module
    if DATA_UPLOAD is not None:
        result = DATA_UPLOAD.run(device_id, timestamp, data_type,
                                 images, data, data_id, resources, options)
    return result

def UploadDataPool(args):
    result = None
    upload_data_pool_module = args[0]
    device_id = args[1]
    timestamp = args[2]
    data_type = args[3]
    images = args[4]
    data = args[5]
    data_id = args[6]
    resources = args[7]
    options = args[8]
    result = UploadDataPoolRun(upload_data_pool_module,
                               device_id, timestamp, data_type,
                               images, data, data_id, resources, options)
    return result

class SoracomDataUploadThread(threading.Thread):
    def __init__(self, useMultiprocessing=True,
                 upload_data_pool_module='', multiprocessingTimeout=10):
        super().__init__()
        self.upload_data_pool_module = upload_data_pool_module
        if useMultiprocessing:
            self.multiprocessingPool = Pool(processes=1)
        else:
            self.multiprocessingPool = None

        self.locker = threading.Lock()
        self.condition = threading.Condition()
        self.conditionWait = False

        self.device_id = None
        self.timestamp = None
        self.data_type = None
        self.images = None
        self.data = None
        self.data_id = None
        self.privacy = False
        self.result = None
        self.working = False
        self.options = None
        self.resources = None
        self.multiprocessingTimeout = multiprocessingTimeout
        # thread demon
        self.daemon = True
        # start thread
        self.start()

    def _startUpload(self, device_id, timestamp, data_type,
                     images, data, data_id, resources, options):
        self.device_id = device_id
        self.timestamp = timestamp
        self.data_type = data_type
        self.images = images
        self.data = data
        self.data_id = data_id
        self.resources = resources
        self.options = options
        self.result = None
        self.working = True
        self.condition.acquire()
        if self.conditionWait:
            self.condition.notify()
        self.condition.release()

    def _getWorking(self):
        return self.working

    def _getResult(self):
        return self.result

    def _resetUpload(self):
        self.device_id = None
        self.timestamp = None
        self.data_type = None
        self.images = None
        self.data = None
        self.data_id = None
        self.privacy = False
        self.result = None

    def resetUpload(self):
        self.locker.acquire()
        try:
            self._resetUpload
        except BaseException:
            self.locker.release()
        self.locker.release()

    def getWorking(self):
        result = None
        self.locker.acquire()
        try:
            result = self._getWorking()
        except BaseException:
            self.locker.release()
            return None
        self.locker.release()
        return result

    def getResult(self):
        result = None
        self.locker.acquire()
        try:
            result = self._getResult()
        except BaseException:
            self.locker.release()
            return None
        self.locker.release()
        return result

    def startUpload(self, device_id, timestamp, data_type,
                    images, data, data_id=-1, resources=None, options={}):
        self.locker.acquire()
        try:
            self._startUpload(device_id, timestamp, data_type,
                              images, data, data_id, resources, options)
        except BaseException:
            self.locker.release()
            return
        self.locker.release()
        return

    def _uploadDataThread(self):
        #Lock
        logger.debug('_uploadDataThread start')
        self.locker.acquire()
        device_id = self.device_id
        timestamp = self.timestamp
        data_type = self.data_type
        images = self.images
        data = self.data
        data_id = self.data_id
        resources = self.resources
        options = self.options
        self.locker.release()
        logger.info('uploadData start %d', data_id)
        if self.multiprocessingPool is not None:
            res = self.multiprocessingPool.apply_async(UploadDataPool, \
                  [(self.upload_data_pool_module, device_id, timestamp,
                    data_type, images, data, data_id, resources, options)])
            try:
                # Wait timeout seconds for func to complete.
                result = res.get(self.multiprocessingTimeout)
            except TimeoutError:
                res.revoke()
                logger.error("id %d is aborted due to timeout", data_id)
                return None
        else:
            result = UploadDataPoolRun(self.upload_data_pool_module, \
                    device_id, timestamp, data_type, images, data, data_id, resources, options)
        logger.info('uploadData end %d', data_id)
        self.locker.acquire()
        self.result = result
        self.locker.release()
        logger.debug('_uploadDataThread end')
        return True

    def _threadMain(self):
        while True:
            try:
                self.condition.acquire()
                self.conditionWait = True
                self.condition.wait(None)
                self.conditionWait = False
                self.condition.release()
                logger.info('uploadData')
                try:
                    self._uploadDataThread()
                except Exception as err:
                    import traceback
                    traceback.print_exc()
                    logger.error(err)
                    sleep(5)
                self.locker.acquire()
                self.working = False
                self.locker.release()
            except Exception as err:
                import traceback
                traceback.print_exc()
                logger.error(err)
                sleep(5)

    def run(self):
        self._threadMain()
