# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0 Sample
# File Name       : soracom_device_interface.py
# Copyright       : 2015-2021 SORACOM, INC.
# ======================================================

import os
import time
from time import sleep
import logging
from logging import getLogger
import threading
import json
import requests
import cv2
import numpy
import traceback
import hashlib

from fisheye_correction import FisheyeCorrection

logger = getLogger(__name__)

class SoracomDeviceInterface(threading.Thread):

    def __init__(self):
        super().__init__()
        self.locker = threading.Lock()
        self.condition = threading.Condition()
        self.conditionWait = False

        self.url = 'http://127.0.0.1:8080'
        self.width = 1280
        self.height = 720
        self.resize = 2
        self.fcf = cv2.CascadeClassifier('./haarcascade_frontalface_alt2.xml')
        self.ecf = cv2.CascadeClassifier('./haarcascade_eye.xml')
        self.capStarted = False
        self.recordingStarted = False
        self.preGetImageBGR = None
        self.preGetImageTime = None
        self.headers = {"Content-Type": "application/json"}

        self.__fcalib = FisheyeCorrection()

        self.__debug_mode = False
        self.__webcam_capture = None

        # thread demon
        self.daemon = True

        # start thread
        self.start()

    def _getCapStarted(self):
        capStarted = False
        self.locker.acquire()
        try:
            capStarted = self.capStarted
        except BaseException:
            self.locker.release()
            return False
        self.locker.release()
        return capStarted

    def _setCapStarted(self, state):
        self.locker.acquire()
        try:
            self.capStarted = state
        except Exception:
            self.locker.release()
            return
        self.locker.release()

    def _getRecordingStarted(self):
        recordingStarted = False
        self.locker.acquire()
        try:
            recordingStarted = self.recordingStarted
        except BaseException:
            self.locker.release()
            return False
        self.locker.release()
        return recordingStarted

    def _setPreGetImage(self, imageBGR):
        self.locker.acquire()
        try:
            self.preGetImageBGR = imageBGR
            self.preGetImageTime = time.time()
            logger.info('self.preGetImageTime: %s', str(self.preGetImageTime))
        except Exception:
            self.preGetImageBGR = None
            self.preGetImageTime = None
        self.locker.release()

    def _getPreGetImage(self):
        preGetImageBGR = None
        self.locker.acquire()
        try:
            if self.capStarted:
                if self.preGetImageBGR is not None:
                    if self.preGetImageTime is not None:
                        now = time.time()
                        logger.info('now: %s', str(now))
                        if now - 60 < self.preGetImageTime:
                            logger.info('preGetImageBGR')
                            preGetImageBGR = self.preGetImageBGR
                    self.preGetImageBGR = None
                    self.preGetImageTime = None
        except Exception:
            self.preGetImageBGR = None
            self.preGetImageTime = None
        self.locker.release()
        return preGetImageBGR

    def _preGetCapture(self):
        self.condition.acquire()
        if self.conditionWait:
            self.condition.notify()
        self.condition.release()

    def setCaptureSize(self, uri, width, height, resize):
        self.url = uri
        self.width = width
        self.height = height
        self.resize = resize

    def setUpdateFCalibPath(self, filename):
        self.__fcalib.update_calibration(filename)

    def setWebpcamDebug(self, camera_num=-1):
        self.__debug_mode = True
        self.__webcam_capture = cv2.VideoCapture(camera_num)

    def webCamRelease(self):
        self.__webcam_capture.release()

    def getWebcamImage(self):
        ret, frame = self.__webcam_capture.read()
        mag_rate_hight = float(self.height / frame.shape[0])
        mag_rate_width = float(self.width / frame.shape[1])
        if mag_rate_width > mag_rate_hight:
            dst = cv2.resize(frame, dsize=None,
                             fx=mag_rate_width, fy=mag_rate_width)
            margin = int((dst.shape[0] - self.height)/2)
            frame = dst[margin: (dst.shape[0] - margin), 0: dst.shape[1]]
        else:
            dst = cv2.resize(frame, dsize=None,
                             fx=mag_rate_hight, fy=mag_rate_hight)
            margin = int((dst.shape[1] - self.width)/2)
            frame = dst[0: dst.shape[0], margin: (dst.shape[1] - margin)]
        return frame

    def startCapture(self):
        logger.debug('startCapture start')
        try:
            if not self._getCapStarted():
                # stop camera
                _ = self.stopCameraRecording()
                response = requests.get(self.url + '/v1/stopCameraCapture')
                logger.info('stopCameraCapture response: %s', str(response))
                if response.status_code != 200:
                    return False
                # wait for reboot
                sleep(5)

                # camera settings
                cameraStateJson = {"width": self.width,
                                   "height": self.height, "framerate": 10}
                logger.info('cameraState cameraStateJson: %s',
                            str(cameraStateJson))
                response = requests.patch(
                    self.url + '/v1/cameraState', json.dumps(cameraStateJson), timeout=5.0)
                logger.info('cameraState response: %s', str(response))
                if response.status_code != 200:
                    return False

                # additional camera settings
                response = requests.get(self.url + '/v1/startCameraCapture')
                logger.info('startCameraCapture response: %s', str(response))
                if response.status_code != 200:
                    return False
                self._setCapStarted(True)

                # wait for initialization
                sleep(3)

            return True
        except Exception as e:
            logger.error(e)
            self._setCapStarted(False)
            return False

    def startCameraRecording(self):
        logger.debug('startCameraRecording start')
        try:
            if not self._getCapStarted():
                self.startCapture()
            response = requests.get(self.url + '/v1/startCameraRecording')
            if response.status_code != 200:
                return False
            return True
        except Exception as e:
            logger.error(e)
            return False

    def stopCameraRecording(self):
        logger.debug('stopCameraRecording start')
        try:
            response = requests.get(self.url + '/v1/stopCameraRecording')
            if response.status_code != 200:
                return False
            return True
        except Exception as e:
            logger.error(e)
            return False


    def getCaptureJpeg(self):
        logger.debug('getCaptureJpeg start')
        try:
            res = self.startCapture()
            if not res:
                return None

            response = requests.get(self.url + '/v1/cameraJpegImage')
            if response.status_code != 200:
                # retry
                self._setCapStarted(False)
                res = self.startCapture()
                if not res:
                    return None
                response = requests.get(self.url + '/v1/cameraJpegImage')
                if response.status_code != 200:
                    return None

            return response.content
        except Exception:
            self._setCapStarted(False)
            return None

    def getCapture(self, preGet=False):
        logger.debug('getCapture start')
        try:
            image_bgr = None
            if self.__debug_mode is False:
                content = self._getPreGetImage()
                if content is None:
                    res = self.startCapture()
                    if not res:
                        logger.info("can't get an image")
                        return None

                    response = requests.get(self.url + '/v1/cameraImage')
                    if response.status_code != 200:
                        self._setCapStarted(False)
                        return None
                    content = response.content

                logger.debug('requests.get end')
                logger.debug('numpy.frombuffer start')
                img_numpy_rgb = numpy.frombuffer(content, numpy.uint8).reshape(
                    self.height, self.width, 3)
                logger.debug('numpy.frombuffer end')

                logger.debug('cv2.cvtColor start')
                image_bgr = cv2.cvtColor(img_numpy_rgb, cv2.COLOR_RGB2BGR)
                logger.debug('cv2.cvtColor end')
            else:
                image_bgr = self.getWebcamImage()

            if self.resize == 1:
                ret_bgr = image_bgr
            else:
                logger.debug('cv2.resize start')
                size = (self.width//self.resize, self.height//self.resize)
                ret_bgr = cv2.resize(image_bgr, size)
                logger.debug('cv2.resize end')
            ret_bgr = self.__fcalib.correction(ret_bgr)

            if preGet:
                self._preGetCapture()
            logger.debug('getCapture end')
            return ret_bgr
        except Exception:
            self._setCapStarted(False)
            return None

    def downloadImage(self, url):
        ret_bgr = None
        file_name = hashlib.sha256(url.encode()).hexdigest() + os.path.basename(url)
        file_path = os.path.join("/tmp", file_name)
        if os.path.exists(file_path):
            logger.debug('read a file from cache')
            ret_bgr = cv2.imread(file_path)
        try:
            logger.debug('downloadCaptureImage start')
            response = requests.get(url, stream=True)
            if response.status_code != 200:
                logger.error("can't download file")
                logger.error(response)
            elif response.status_code == 200:
                logger.debug('requests.get end')
                image = numpy.asarray(bytearray(response.raw.read()), dtype="uint8")
                image_bgr = cv2.imdecode(image,  cv2.IMREAD_COLOR)
                logger.debug('cv2 resize start')
                image_bgr = cv2.resize(image_bgr,(
                    self.width, self.height))
                logger.debug('cv2 resize end')
                cv2.imwrite(file_path, image_bgr)
                if self.resize == 1:
                    ret_bgr = image_bgr
                else:
                    logger.debug('cv2.resize start')
                    size = (self.width//self.resize, self.height//self.resize)
                    ret_bgr = cv2.resize(image_bgr, size)
                    logger.debug("resize: {}".format(self.resize))
                    logger.debug('cv2.resize end')
                    ret_bgr = self.__fcalib.correction(ret_bgr)
                    logger.debug('downloadCaptureImage end')
        except Exception as err:
            logger.error(traceback.format_exc())
            logger.error("can't return image by %s", err)
            return None
        return ret_bgr

    def findFaces(self, image_bgr):
        "return face outlines"

        face_list = []
        logger.debug('findFaces start')
        if image_bgr is None:
            return face_list
        img_gray = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2GRAY)
        face_list = self.fcf.detectMultiScale(
            img_gray,
            minSize=(30, 30))
        logger.debug("%d faces are found.", len(face_list))
        return face_list

    def findEyes(self, image_bgr):
        "return eye outlines"

        face_list = []
        if image_bgr is None:
            return face_list
        img_gray = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2GRAY)
        eye_list = self.ecf.detectMultiScale(
            img_gray,
            minSize=(30, 30))

        return eye_list

    def getCaptureFace(self, privacy=False, preGet=False):
        "return a image with a face list"

        face_info = {}
        face_list = numpy.array([])
        logger.debug('getCaptureFace start')
        image_bgr = self.getCapture(preGet)
        if image_bgr is None:
            face_info['image'] = None
            face_info['face_list'] = face_list
            return face_info
        if privacy:
            eye_lists = self.findEyes(image_bgr)
            for eye_list in eye_lists:
                x1 = int(eye_list[0])
                x2 = int(eye_list[0] + eye_list[2])
                y1 = int(eye_list[1])
                y2 = int(eye_list[1] + eye_list[3])

                # make eye area smaller
                small = cv2.resize(image_bgr[y1:y2, x1:x2], (5, 5))
                # expand the image to make mosaic
                zoom = cv2.resize(small, ((x2 - x1), (y2 - y1)),
                                  interpolation=cv2.INTER_AREA)
                image_bgr[y1:y2, x1:x2] = zoom
            face_list = eye_lists
        else:
            face_list = self.findFaces(image_bgr)
        face_info['image'] = image_bgr
        face_info['face_list'] = face_list
        logger.debug('getCaptureFace end')
        return face_info

    def getCameraState(self):
        logger.info('getCameraState start')
        try:
            response = requests.get(self.url + '/v1/cameraState')
            if response.status_code != 200:
                return None
            js = json.loads(response.content.decode('utf-8'))
            return js
        except Exception:
            return None

    def setCameraState(self, cameraState):
        logger.info('setCameraState cameraState: %s', str(cameraState))
        try:
            response = requests.patch(
                self.url + '/v1/cameraState',
                json.dumps(cameraState),
                headers=self.headers)
            if response.status_code != 200:
                return False
            return True
        except Exception:
            return False

    def getDeviceState(self):
        logger.info('getDeviceState start')
        try:
            response = requests.get(self.url + '/v1/deviceState')
            logger.info('getDeviceState response.status_code: %d',
                        response.status_code)
            logger.debug('getDeviceState response.content: %s',
                         response.content)
            if response.status_code == 200:
                js = json.loads(response.content.decode('utf-8'))
                return js
            return None
        except Exception:
            return None

    def _threadMain(self):

        while True:
            try:
                self.condition.acquire()
                self.conditionWait = True
                self.condition.wait(None)
                self.conditionWait = False
                self.condition.release()
                if self._getCapStarted():
                    logger.info('preGet')
                    response = requests.get(self.url + '/v1/cameraImage')
                    if response.status_code == 200:
                        self._setPreGetImage(response.content)
            except Exception:
                sleep(5)

    def run(self):
        self._threadMain()
