# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0 Sample
# File Name       : soracom_image_analize_object_detection.py
# Copyright       : 2015-2020 SORACOM, INC.
# ======================================================

# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#  git clone https://github.com/google-coral/examples-camera.git --depth 1

import glob
import os
import sys
import json
import time
import collections
from logging import getLogger
import re
import traceback
import platform
import cv2
import numpy
from PIL import Image
import numpy as np
import tflite_runtime.interpreter as tflite
from soracom_image_util import SoracomImageUtil

logger = getLogger(__name__)

COLOR_PURPLE = "\033[35m"
COLOR_RESET = "\033[0m"

IMAGE_UTIL = SoracomImageUtil()

FREE_PARAM = str(os.getenv('SORACOM_ENV_FREE_PARAM', '{}'))

try:
    FREE_PARAM_OBJECT = json.loads(FREE_PARAM)
except Exception:
    FREE_PARAM_OBJECT = {}
    logger.error("free param's json load error")

class SoracomImageAnalizeObjectDetection():
    def __init__(self):
        super().__init__()
        SORACOM_ENV_APP_DATA_ROOT = os.getenv('SORACOM_ENV_APP_DATA_ROOT', '/opt/app/')
        #Please set the same name as the IMAGE_ANALIZE_DIR_NAME of PreSetup.
        IMAGE_ANALIZE_DIR_NAME = 'soracom_image_analize_model'
        movenet_model_name = FREE_PARAM_OBJECT.get(
            'Model', 'movenet_multipose_lightning.tflite')
        movenet_model_path = os.path.join(SORACOM_ENV_APP_DATA_ROOT, IMAGE_ANALIZE_DIR_NAME,\
                            movenet_model_name)
        try:
            logger.debug("load movenet model at: %s", movenet_model_path)
            self.movenet_interpreter = tflite.Interpreter(movenet_model_path)
            logger.debug("[Success] Loaded Movenet Model")
        except Exception as error:
            traceback.print_exc()
            logger.error("Can't create movenet interpreter by: %s", error)
            sys.exit(1)

        self.input_size = (256, 256)
        # get threshhold for the confidence
        self.bbox_threshold = FREE_PARAM_OBJECT.get('OThresh', 0.2)
        self.keypoint_threshold = 0.2

    def saveFile(self, image, filename="object.jpg", note=''):
        _, im_data = cv2.imencode('.jpeg', image)
        fullpath = os.path.join('/tmp', filename)
        try:
            logger.info('image file: %s note: %s', fullpath, note)
            with open(fullpath, mode='wb') as fout:
                fout.write(im_data)
        except FileExistsError as error:
            raise error

    def predictPose(self, image, input_size=(256, 256)):
        input_tensor = image.copy()
        input_tensor = cv2.cvtColor(input_tensor, cv2.COLOR_BGR2RGB)
        input_tensor = cv2.resize(input_tensor, dsize=input_size)
        input_tensor = np.expand_dims(input_tensor, 0)

        input_details = self.movenet_interpreter.get_input_details()
        output_details = self.movenet_interpreter.get_output_details()

        input_tensor_index = input_details[0]['index']
        input_shape = input_tensor.shape
        self.movenet_interpreter.resize_tensor_input(input_tensor_index, input_shape)
        self.movenet_interpreter.allocate_tensors()

        self.movenet_interpreter.set_tensor(input_tensor_index, input_tensor)

        self.movenet_interpreter.invoke()

        keypoints_with_scores = self.movenet_interpreter.get_tensor(output_details[0]['index'])
        keypoints_with_scores = np.squeeze(keypoints_with_scores)

        return keypoints_with_scores

    def analizeImage(self, analize_type, image, image_id=-1, resources=None, options={}):
        logger.debug('start analizeImage analize_type: %s image_id: %d resources: %s \
                     options: %s', analize_type, image_id, resources, options)
        debug_mode = options.get('debug', False)
        photography_time = options.get('photography_time', None)

        results = {}
        start_ms = time.time()
        logger.info("start analizeImage at: %s", start_ms)

        keypoints_with_scores = self.predictPose(image, input_size=self.input_size)

        logger.debug("image.shape: %s", image.shape)
        keypoints_list, scores_list, bbox_list = self.getDetailLists(keypoints_with_scores, image.shape)

        if len(keypoints_list) == 0:
            return None

        inference_time = (time.time() - start_ms)
        cv2_im = self.drawDetectionOnImage(keypoints_list, scores_list, bbox_list, image)
        results['image'] = cv2_im

        if debug_mode:
            self.saveFile(cv2_im)

        # results['objects'] = objs
        results['objects'] = [{}]
        results['options'] = options
        results['image_id'] = image_id
        results['photography_time'] = photography_time

        logger.info("end analizeImage at: %s elapsed time: %s", time.time(), inference_time)

        return results

    def getDetailLists(self, keypoints_with_scores, size):
        image_height, image_width = size[:2]
        logger.debug("image_height: %s", image_height)
        logger.debug("image_width:  %s", image_width)
        keypoints_list, scores_list, bbox_list = [], [], []
        for kp in keypoints_with_scores:
            bbox_score = kp[55]
            if bbox_score < self.bbox_threshold:
                continue

            keypoints = []
            scores = []
            for index in range(17):
                kp_x = int(image_width * kp[index*3+1])
                kp_y = int(image_height * kp[index*3+0])
                score = kp[index*3+2]
                keypoints.append([kp_x, kp_y])
                scores.append(score)
            bbox_ymin = int(image_height * kp[51])
            bbox_xmin = int(image_width * kp[52])
            bbox_ymax = int(image_height * kp[53])
            bbox_xmax = int(image_width * kp[54])

            keypoints_list.append(keypoints)
            scores_list.append(scores)
            bbox_list.append([bbox_xmin, bbox_ymin, bbox_xmax, bbox_ymax, bbox_score])

        return keypoints_list, scores_list, bbox_list

    def drawDetectionOnImage(self, keypoints_list, scores_list, bbox_list, img):
        render = img.copy()
        for (keypoints, scores, bbox) in zip(keypoints_list, scores_list, bbox_list):
            cv2.rectangle(render, (bbox[0], bbox[1]), (bbox[2], bbox[3]), (0,255,0), 2)

            kp_links = [
                (0,1),(0,2),(1,3),(2,4),(0,5),(0,6),(5,6),(5,7),(7,9),(6,8),
                (8,10),(11,12),(5,11),(11,13),(13,15),(6,12),(12,14),(14,16)
            ]
            for kp_idx_1, kp_idx_2 in kp_links:
                kp_1 = keypoints[kp_idx_1]
                kp_2 = keypoints[kp_idx_2]
                score_1 = scores[kp_idx_1]
                score_2 = scores[kp_idx_2]
                if score_1 > self.keypoint_threshold and score_2 > self.keypoint_threshold:
                    cv2.line(render, tuple(kp_1), tuple(kp_2), (0,0,255), 2)

            for (keypoint, score) in zip(keypoints, scores):
                if score > self.keypoint_threshold:
                    cv2.circle(render, tuple(keypoint), 4, (0,0,255), -1)

        return render

SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION = None

def run(analize_type, image, image_id, resources, options):
    global SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION
    if SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION is None:
        SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION = SoracomImageAnalizeObjectDetection()
    result = SORACOM_IMAGE_ANALIZE_OBJECT_DETECTION.analizeImage(
        analize_type, image, image_id, resources, options)
    return result
