# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0
# File Name       : soracom_image_analize_thread.py
# Copyright       : 2015-2020 SORACOM, INC.
# ======================================================

from time import sleep
import threading
from multiprocessing import Pool
import importlib

from logging import getLogger

logger = getLogger(__name__)

IMAGE_ANALIZE = None
IMAGE_ANALIZE_MODULE = ''

def AnalizeImagePoolRun(analize_image_pool_module,
                        analize_type, image, image_id, resources, options):
    global IMAGE_ANALIZE
    global IMAGE_ANALIZE_MODULE
    result = None
    logger.debug("IMAGE_ANALIZE_MODULE is %s.", IMAGE_ANALIZE_MODULE)
    if analize_image_pool_module is not None and len(analize_image_pool_module) > 0:
        if IMAGE_ANALIZE_MODULE != analize_image_pool_module:
            logger.info("import image analize module name is %s.", analize_image_pool_module)
            IMAGE_ANALIZE = importlib.import_module(analize_image_pool_module)
            if IMAGE_ANALIZE is not None:
                IMAGE_ANALIZE_MODULE = analize_image_pool_module
    if IMAGE_ANALIZE is not None:
        result = IMAGE_ANALIZE.run(analize_type, image, image_id, resources, options)
    return result

def AnalizeImagePool(args):
    result = None
    analize_image_pool_module = args[0]
    analize_type = args[1]
    image = args[2]
    image_id = args[3]
    resources = args[4]
    options = args[5]
    result = AnalizeImagePoolRun(analize_image_pool_module, \
                  analize_type, image, image_id, resources, options)
    return result

class SoracomImageAnalizeThread(threading.Thread):
    def __init__(self, useMultiprocessing=True, analize_image_pool_module='',
                 multiprocessing_timeout=10):
        super().__init__()
        self.analize_image_pool_module = analize_image_pool_module
        if useMultiprocessing:
            self.multiprocessing_pool = Pool(processes=1)
        else:
            self.multiprocessing_pool = None

        self.locker = threading.Lock()
        self.condition = threading.Condition()
        self.condition_wait = False

        self.analize_type = None
        self.image = None
        self.image_id = None
        self.privacy = False
        self.result = None
        self.working = False
        self.options = None
        self.resources = None
        self.multiprocessing_timeout = multiprocessing_timeout

        # thread demon
        self.daemon = True

        # start thread
        self.start()

    def _startAnalize(self, analize_type, image, image_id, resources, options):
        self.analize_type = analize_type
        self.image = image
        self.image_id = image_id
        self.resources = resources
        self.options = options
        self.result = None
        self.working = True
        self.condition.acquire()
        if self.condition_wait:
            self.condition.notify()
        self.condition.release()

    def _getWorking(self):
        return self.working

    def _getResult(self):
        return self.result

    def _resetAnalize(self):
        self.analize_type = None
        self.image = None
        self.image_id = None
        self.privacy = False
        self.result = None

    def resetAnalize(self):
        self.locker.acquire()
        try:
            self._resetAnalize
        except BaseException:
            self.locker.release()
        self.locker.release()

    def getWorking(self):
        result = None
        self.locker.acquire()
        try:
            result = self._getWorking()
        except BaseException:
            self.locker.release()
            return None
        self.locker.release()
        return result

    def getResult(self):
        result = None
        self.locker.acquire()
        try:
            result = self._getResult()
        except BaseException:
            self.locker.release()
            return None
        self.locker.release()
        return result

    def startAnalize(self, analize_type, image, image_id=-1, resources=None, options={}):
        self.locker.acquire()
        try:
            self._startAnalize(analize_type, image, image_id, resources, options)
        except BaseException:
            self.locker.release()
            return
        self.locker.release()
        return

    def _analizeImageThread(self):
        #Lock
        logger.debug('_analizeImageThread start')
        self.locker.acquire()
        analize_type = self.analize_type
        image = self.image
        image_id = self.image_id
        resources = self.resources
        options = self.options
        self.locker.release()
        logger.info('analizeImage start for image id: %d', image_id)
        if self.multiprocessing_pool is not None:
            res = self.multiprocessing_pool.apply_async(AnalizeImagePool, \
                [(self.analize_image_pool_module, analize_type, image,
                  image_id, resources, options)])
            try:
                # Wait timeout seconds for func to complete.
                result = res.get(self.multiprocessing_timeout)
            except TimeoutError:
                res.revoke()
                logger.error("id %d is aborted due to timeout", image_id)
                return None
        else:
            result = AnalizeImagePoolRun(self.analize_image_pool_module, \
                    analize_type, image, image_id, resources, options)
        logger.info('analizeImage end %d', image_id)
        self.locker.acquire()
        self.result = result
        self.locker.release()
        logger.debug('_analizeImageThread end')
        return True

    def _threadMain(self):
        while True:
            try:
                self.condition.acquire()
                self.condition_wait = True
                self.condition.wait(None)
                self.condition_wait = False
                self.condition.release()
                logger.info('analizeImage')
                try:
                    self._analizeImageThread()
                except Exception as err:
                    import traceback
                    traceback.print_exc()
                    logger.error(err)
                    sleep(5)
                self.locker.acquire()
                self.working = False
                self.locker.release()
            except Exception as err:
                import traceback
                traceback.print_exc()
                logger.error(err)
                sleep(5)

    def run(self):
        self._threadMain()
