# ======================================================
# Project Name    : SORACOM Mosaic CameraApp0 Sample
# File Name       : soracom_image_util.py
# Copyright       : 2015-2020 SORACOM, INC.
# ======================================================

import os
from logging import getLogger
import cv2
import numpy

logger = getLogger(__name__)


class SoracomImageUtil:
    def __init__(self, gamma=0, gamma_mul=1.0):
        self.__gamma = None
        if gamma != 0:
            self.setImageR_Gamma(gamma, gamma_mul)

    def centerAspectBox(self, image, fill_color=255):
        """
        The image size is changed to square while maintaining the aspect ratio of the original image.
        The margin is filled with the color specified by the argument.

        A landscape image is converted to a square image with
        a horizontal length and top and bottom margins.
        Also, a vertically long image is converted to
        an image that has left and right margins in a square image that is vertically long.

        Parameters
        ----------
        image: numpy.ndarray
        　　　　Original image.
        fillColor: numpy.ndarray
        　　　　　　　back color, set  gray scale.

        Returns
        -------
        image : numpy.ndarray
                Square image.
        fit_size : int
                Length of a square piece.
        offset_x : int
                Image location information x.
        offset_y : int
                Image location information y.
        """

        h, w = image.shape[:2]
        if h > w:
            fit_size = h
            limit = w
        else:
            fit_size = w
            limit = h
        start = int((fit_size - limit) / 2)
        fin = int((fit_size + limit) / 2)
        new_img = cv2.resize(numpy.zeros(
            (1, 1, 3), numpy.uint8), (fit_size, fit_size))
        new_img.fill(fill_color)
        offset_x = 0
        offset_y = 0
        if fit_size == h:
            offset_x = start
            new_img[:, start:fin] = image
        else:
            offset_y = start
            new_img[start:fin, :] = image
        return new_img, fit_size, offset_x, offset_y


    def cropImage(self, image, t, b, l, r):
        """
        Crop the image from the top, bottom, left and right.

        Parameters
        ----------
        image: numpy.ndarray
                Image to cut out image.
        t : int
            Size to cut from top.
        b : int
            Size to cut from bottom.
        l : int
            Size to cut from left.
        r : int
            size to cut from right.

        Returns
        -------
        image: numpy.ndarray
            　　Processed image.
        """

        h, w = image.shape[0:2]
        return image[t:h-b, l:w-r]

    def imageR_Gamma(self, image):
        """
        Performs redness correction.

        Parameters
        ----------
        image : numpy.ndarray
                Original image.

        Returns
        -------
        image : numpy.ndarray
                Processed image.
        """

        red_matrix = image[:, :, 2]
        red_matrix = red_matrix.astype(numpy.float32)
        red_matrix = numpy.power(red_matrix / 255, 1.0 / self.__gamma)
        red_matrix *= 255 * self.__gamma_mul
        red_matrix = red_matrix.astype(numpy.uint8)
        image[:, :, 2] = red_matrix
        return image

    def imageHSV(self, image):
        """
        Convert the image color space to HSV.

        Parameters
        ----------
        image : numpy.ndarray
                Original image.

        Returns
        -------
        s : int
            Value of S in HSV color space
        v : int
            Value of V in HSV color space
        """

        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        s = (hsv.T[1].flatten().mean())
        v = (hsv.T[2].flatten().mean())
        return s, v

    def makeThumbnailsSplite(
            self, images, thumbnail_width=90, thumbnail_hight=90,
            w_count=5, h_count=5, back_color=255):
        """
        Create a sprite image with thumbnail images arranged in one image.

        Parameters
        ----------
        images:　list
        　　　　　list of thumbnail images.
        thumbnail_width:　int
        　　　　　　　　　　 thumbnail image width.
        thumbnail_hight : int
                          thumbnail image thumbnail_hight.
        w_count : int
                  number of thumbnail images in the horizontal direction.
        h_count : int
                  number of thumbnail images in the vertical direction.
        back_color : int
                     back color, set  gray scale.

        Returns
        -------
        image: numpy.ndarray
               Sprite image.
        """

        background = numpy.zeros(
            (thumbnail_hight*h_count, thumbnail_width*w_count, 3),
            dtype=numpy.uint8)
        background.fill(back_color)

        for index, image in enumerate(images):
            if index >= w_count*h_count:
                logger.error("over 25 images during this 1 min!")
                break
            new_img, fit_size, offset_x, offset_y = self.centerAspectBox(image, back_color)
            thumbnail_image = cv2.resize(
                new_img, (thumbnail_width, thumbnail_hight))
            pos_y = int(index / w_count)*thumbnail_hight
            pos_x = (index % w_count)*thumbnail_width
            background[pos_y:(thumbnail_image.shape[0] + pos_y),
                       pos_x:(thumbnail_image.shape[1] + pos_x)] = thumbnail_image

        return background

    def removeoise(self, image):
        """
        remove noise from the image.

        Parameters
        ----------
        images: numpy.ndarray
                Original image.
        Returns
        -------
        image: numpy.ndarray
               Image with noise removal.
        """

        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        filtered = cv2.adaptiveThreshold(image.astype(
            numpy.uint8), 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 9, 41)
        kernel = numpy.ones((1, 1), numpy.uint8)
        opening = cv2.morphologyEx(filtered, cv2.MORPH_OPEN, kernel)
        closing = cv2.morphologyEx(opening, cv2.MORPH_CLOSE, kernel)
        image = self.imageSmoothening(image)
        or_image = cv2.bitwise_or(image, closing)
        return or_image

    def imageSmoothening(self, image):
        """
        smoothing the image.

        Parameters
        ----------
        images: numpy.ndarray
                Original image.
        Returns
        -------
        image: numpy.ndarray
               Image with noise removal.
        """
        ret1, th1 = cv2.threshold(
            image, BINARY_THREHOLD, 255, cv2.THRESH_BINARY)
        ret2, th2 = cv2.threshold(
            th1, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
        blur = cv2.GaussianBlur(th2, (1, 1), 0)
        ret3, th3 = cv2.threshold(
            blur, 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)
        return th3

    def rotateImage(self, image):
        """
        Correct text skew of the input image.

        Parameters
        ----------
        images: numpy.ndarray
                Original image.

        Returns
        -------
        rotate_image: numpy.ndarray
                      rotaled image.
        """
        gray_image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        gray_image = cv2.bitwise_not(gray_image)
        gray_image = cv2.GaussianBlur(gray_image, (5, 5), 0)
        thresh = cv2.threshold(gray_image, 0, 255,
                               cv2.THRESH_BINARY | cv2.THRESH_OTSU)[1]
        coords = numpy.column_stack(numpy.where(thresh > 0))
        angle = cv2.minAreaRect(coords)[-1]
        if angle < -45:
            angle = -(90 + angle)
        else:
            angle = -angle
        (h, w) = image.shape[:2]
        center = (w // 2, h // 2)
        M = cv2.getRotationMatrix2D(center, angle, 1.0)
        rotate_image = cv2.warpAffine(image, M, (w, h),
                                      flags=cv2.INTER_CUBIC,
                                      borderMode=cv2.BORDER_REPLICATE)
        return rotate_image


    def save(self, image, directory, filename):
        """
        Save the image data to Disk.

        Parameters
        ----------
        image : numpy.ndarray
                Original image.
        directory : str
                    path of directory.
        filename : str
                file name.

        Returns
        -------
        result : bool
                 Processing result
        """

        if image is None:
            return False
        fullpath = os.path.join(directory, filename)
        try:
            logger.info('save image to %s', fullpath)
            if image is not None:
                with open(fullpath, mode='wb') as fout:
                    fout.write(image)
        except FileExistsError as error:
            raise error
        return True

    def saveJPEG(self, image, directory, filename, quality=75):
        """
        Convert image data to jpeg and save to Disk.

        Parameters
        ----------
        image : numpy.ndarray
                Original image.
        directory : str
                    path of directory.
        filename : str
                   file name.

        Returns
        -------
        result : bool
                 Processing result
        """

        if image is None:
            return False
        fullpath = os.path.join(directory, filename)
        result = False
        try:
            encode_param = [cv2.IMWRITE_JPEG_QUALITY, quality]
            logger.info('save image to %s', fullpath)
            result = cv2.imwrite(fullpath, image, encode_param)
        except FileExistsError as error:
            raise error
        return result

    def setColorBalance(self, image, deviceInterface, s_min=50, s_max=150, v_min=100, v_max=200):
        """
        A sample that changes camera parameters based on the HSV value of the image
        If you want to capture the best image for the algorithm,
        change the camera parameters with reference to this sample.

        Parameters
        ----------
        image : numpy.ndarray
                Original image.
        s_min: Minimum change in S value.
        s_max: Maximum change in S value.
        v_min: Minimum change in V value.
        v_max: Maximum change in V value.
        """

        image_s, image_v = self.imageHSV(image)
        logger.info('image_s: %s, image_v: %s', str(image_s), str(image_v))

        contrast = 0
        brightness = 0
        if image_s < s_min:
            contrast += 10
        elif image_s > s_max:
            contrast -= 10
        if image_v < v_min:
            brightness += 10
        elif image_v > v_max:
            brightness -= 10

        if contrast != 0 or brightness != 0:
            cameraState = deviceInterface.getCameraState()
            if cameraState is not None:
                logger.info('get cameraState=%s', str(cameraState))
                contrast = cameraState["contrast"] + contrast
                brightness = cameraState["brightness"] + brightness

                if contrast < -100:
                    contrast = -100
                elif contrast > 100:
                    contrast = 100

                if brightness < -100:
                    brightness = -100
                elif brightness > 100:
                    brightness = 100

                cameraState = {"contrast": contrast, "brightness": brightness}
                logger.info('set cameraState: %s', str(cameraState))
                deviceInterface.setCameraState(cameraState)


    def setImageR_Gamma(self, gamma, gamma_mul):
        """
        Set the gamma value and magnification for redness correction.
        （This value can also be specified in the class constructor.）

        Parameters
        ----------
        gamma : float
                gamma
        gamma_mul : float
                    magnification
        """

        if gamma == 0:
            gamma = 1.0
        self.__gamma = gamma
        self.__gamma_mul = gamma_mul

    def sliceImage(self, image, rect):
        """
        slice the image by specifying the coordinates.
        Rect can be obtained from findFaces and findEyes from
        Soracom Device Interface.

        Parameters
        ----------
        image : numpy.ndarray
                Image to cut out image.
        rect : list
               Size to cut rect(x, y, width, hight).

        Returns
        -------
        image : numpy.ndarray
                Processed image.
        """

        l = rect[0]
        r = rect[0] + rect[2]
        t = rect[1]
        b = rect[1] + rect[3]

        vm = rect[3]
        hm = rect[2]

        h, w = image.shape[0:2]

        l = max(l - hm, 0)
        r = min(r + hm, w)
        t = max(t - vm, 0)
        b = min(b + vm, h)

        return image[t:b, l:r]

    def crop_im(self, image, t, b, l, r):
        return self.cropImage(image, t, b, l, r)

    def set_cb(self, image, deviceInterface):
        return self.setColorBalance(image, deviceInterface)

    def set_imageR_Gamma(self, gamma, gamma_mul):
        return self.setImageR_Gamma(gamma, gamma_mul)

    def slice_im(self, image, rect):
        return self.sliceImage(image, rect)
